MessageNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message'
StructureSpecificNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1' \
                      '/metadata/structurespecific} '
QueryNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/query'
RegistryNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/registry'
StructureNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure'
GenericNS = 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/data/generic'
XMLInstanceNS = 'http://www.w3.org/2001/XMLSchema-instance'

# Definitions for the abbreviations of the namespaces

messageAbbr = 'mes'
genericAbbr = 'gen'
commonAbbr = 'com'
structureAbbr = 'str'
structureSpecificAbbr = 'ss'

Data_Types_VTL = {
    "String": "String",
    "Alpha": "String",
    "AlphaNumeric": "String",
    "Numeric": "String",
    "BigInteger": "Integer",
    "Integer": "Integer",
    "Long": "Integer",
    "Short": "Integer",
    "Decimal": "Number",
    "Float": "Number",
    "Double": "Number",
    "Boolean": "Boolean",
    "URI": "String",
    "Count": "Integer",
    "InclusiveValueRange": "Number",
    "ExclusiveValueRange": "Number",
    "Incremental": "Number",
    "ObservationalTimePeriod": "Date",
    "StandardTimePeriod": "Date",
    "BasicTimePeriod": "Date",
    "GregorianTimePeriod": "Date",
    "GregorianYear": "Date",
    "GregorianYearMonth": "Date",
    "GregorianMonth": "Date",
    "GregorianDay": "Date",
    "ReportingTimePeriod": "TimePeriod",
    "ReportingYear": "TimePeriod",
    "ReportingSemester": "TimePeriod",
    "ReportingTrimester": "TimePeriod",
    "ReportingQuarter": "TimePeriod",
    "ReportingMonth": "TimePeriod",
    "ReportingWeek": "TimePeriod",
    "ReportingDay": "TimePeriod",
    "DateTime": "Date",
    "TimeRange": "Time",
    "Month": "String",
    "MonthDay": "String",
    "Day": "String",
    "Time": "String",
    "Duration": "Duration"
}

Locale_Codes = {
    'ab': 'Abkhaz',
    'aa': 'Afar',
    'af': 'Afrikaans',
    'ak': 'Akan',
    'sq': 'Albanian',
    'am': 'Amharic',
    'ar': 'Arabic',
    'an': 'Aragonese',
    'hy': 'Armenian',
    'as': 'Assamese',
    'av': 'Avaric',
    'ae': 'Avestan',
    'ay': 'Aymara',
    'az': 'Azerbaijani',
    'bm': 'Bambara',
    'ba': 'Bashkir',
    'eu': 'Basque',
    'be': 'Belarusian',
    'bn': 'Bengali',
    'bh': 'Bihari',
    'bi': 'Bislama',
    'bs': 'Bosnian',
    'br': 'Breton',
    'bg': 'Bulgarian',
    'my': 'Burmese',
    'ca': 'Catalan; Valencian',
    'ch': 'Chamorro',
    'ce': 'Chechen',
    'ny': 'Chichewa; Chewa; Nyanja',
    'zh': 'Chinese',
    'cv': 'Chuvash',
    'kw': 'Cornish',
    'co': 'Corsican',
    'cr': 'Cree',
    'hr': 'Croatian',
    'cs': 'Czech',
    'da': 'Danish',
    'dv': 'Divehi; Maldivian;',
    'nl': 'Dutch',
    'dz': 'Dzongkha',
    'en': 'English',
    'eo': 'Esperanto',
    'et': 'Estonian',
    'ee': 'Ewe',
    'fo': 'Faroese',
    'fj': 'Fijian',
    'fi': 'Finnish',
    'fr': 'French',
    'ff': 'Fula',
    'gl': 'Galician',
    'ka': 'Georgian',
    'de': 'German',
    'el': 'Greek, Modern',
    'gn': 'Guaraní',
    'gu': 'Gujarati',
    'ht': 'Haitian',
    'ha': 'Hausa',
    'he': 'Hebrew modern',
    'hz': 'Herero',
    'hi': 'Hindi',
    'ho': 'Hiri Motu',
    'hu': 'Hungarian',
    'ia': 'Interlingua',
    'id': 'Indonesian',
    'ie': 'Interlingue',
    'ga': 'Irish',
    'ig': 'Igbo',
    'ik': 'Inupiaq',
    'io': 'Ido',
    'is': 'Icelandic',
    'it': 'Italian',
    'iu': 'Inuktitut',
    'ja': 'Japanese',
    'jv': 'Javanese',
    'kl': 'Kalaallisut',
    'kn': 'Kannada',
    'kr': 'Kanuri',
    'ks': 'Kashmiri',
    'kk': 'Kazakh',
    'km': 'Khmer',
    'ki': 'Kikuyu, Gikuyu',
    'rw': 'Kinyarwanda',
    'ky': 'Kirghiz, Kyrgyz',
    'kv': 'Komi',
    'kg': 'Kongo',
    'ko': 'Korean',
    'ku': 'Kurdish',
    'kj': 'Kwanyama, Kuanyama',
    'la': 'Latin',
    'lb': 'Luxembourgish',
    'lg': 'Luganda',
    'li': 'Limburgish',
    'ln': 'Lingala',
    'lo': 'Lao',
    'lt': 'Lithuanian',
    'lu': 'Luba-Katanga',
    'lv': 'Latvian',
    'gv': 'Manx',
    'mk': 'Macedonian',
    'mg': 'Malagasy',
    'ms': 'Malay',
    'ml': 'Malayalam',
    'mt': 'Maltese',
    'mi': 'Māori',
    'mr': 'Marathi Marāṭhī',
    'mh': 'Marshallese',
    'mn': 'Mongolian',
    'na': 'Nauru',
    'nv': 'Navajo, Navaho',
    'nb': 'Norwegian Bokmål',
    'nd': 'North Ndebele',
    'ne': 'Nepali',
    'ng': 'Ndonga',
    'nn': 'Norwegian Nynorsk',
    'no': 'Norwegian',
    'ii': 'Nuosu',
    'nr': 'South Ndebele',
    'oc': 'Occitan',
    'oj': 'Ojibwe, Ojibwa',
    'cu': 'Old Church Slavonic',
    'om': 'Oromo',
    'or': 'Oriya',
    'os': 'Ossetian, Ossetic',
    'pa': 'Panjabi, Punjabi',
    'pi': 'Pāli',
    'fa': 'Persian',
    'pl': 'Polish',
    'ps': 'Pashto, Pushto',
    'pt': 'Portuguese',
    'qu': 'Quechua',
    'rm': 'Romansh',
    'rn': 'Kirundi',
    'ro': 'Romanian, Moldavan',
    'ru': 'Russian',
    'sa': 'Sanskrit Saṁskṛta',
    'sc': 'Sardinian',
    'sd': 'Sindhi',
    'se': 'Northern Sami',
    'sm': 'Samoan',
    'sg': 'Sango',
    'sr': 'Serbian',
    'gd': 'Scottish Gaelic',
    'sn': 'Shona',
    'si': 'Sinhala, Sinhalese',
    'sk': 'Slovak',
    'sl': 'Slovene',
    'so': 'Somali',
    'st': 'Southern Sotho',
    'es': 'Spanish; Castilian',
    'su': 'Sundanese',
    'sw': 'Swahili',
    'ss': 'Swati',
    'sv': 'Swedish',
    'ta': 'Tamil',
    'te': 'Telugu',
    'tg': 'Tajik',
    'th': 'Thai',
    'ti': 'Tigrinya',
    'bo': 'Tibetan',
    'tk': 'Turkmen',
    'tl': 'Tagalog',
    'tn': 'Tswana',
    'to': 'Tonga',
    'tr': 'Turkish',
    'ts': 'Tsonga',
    'tt': 'Tatar',
    'tw': 'Twi',
    'ty': 'Tahitian',
    'ug': 'Uighur, Uyghur',
    'uk': 'Ukrainian',
    'ur': 'Urdu',
    'uz': 'Uzbek',
    've': 'Venda',
    'vi': 'Vietnamese',
    'vo': 'Volapük',
    'wa': 'Walloon',
    'cy': 'Welsh',
    'wo': 'Wolof',
    'fy': 'Western Frisian',
    'xh': 'Xhosa',
    'yi': 'Yiddish',
    'yo': 'Yoruba',
    'za': 'Zhuang, Chuang',
    'zu': 'Zulu',
}



import setuptools

with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='SDMXthon',
    description='Library with SDMX to Pandas, Pandas to SDMX, '
                'SDMX validation and SDMX metadata validation',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=setuptools.find_packages(),
    include_package_data=True,
    version='1.1.1',
    license='Apache 2.0',
    license_files='license.txt',
    author='MeaningfulData',
    author_email='javier.hernandez@meaningfuldata.eu',
    project_urls={
        'Bug Tracker': 'https://github.com/Meaningful-Data/sdmxthon/issues',
        'Documentation': 'https://docs.sdmxthon.meaningfuldata.eu',
        'Source Code': 'https://github.com/Meaningful-Data/sdmxthon',
        'Changelog': 'https://github.com/Meaningful-Data/sdmxthon/blob'
                     '/master/Changelog.rst '
    },
    keywords=['SDMX', 'Pandas', 'Validation'],
    install_requires=[
        'lxml',
        'pandas',
        'numpy',
        'validators',
        'requests'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)



import re
from datetime import datetime

import numpy as np
import pandas as pd
from pandas import DataFrame

from sdmxthon.model.definitions import DataStructureDefinition

error_level_facets = 'WARNING'


def get_codelist_values(dsd: DataStructureDefinition) -> dict:
    data = {}

    if dsd.dimension_descriptor.components is not None:

        for element in dsd.dimension_descriptor.components.values():
            if (element.representation is not None and
                    element.representation.codelist is not None and
                    not isinstance(element.representation.codelist, str)):
                data[element.id] = list(element.representation.codelist.
                                        items.keys())

    if (dsd.attribute_descriptor is not None and
            dsd.attribute_descriptor.components is not None):
        for record in dsd.attribute_descriptor.components.values():
            if (record.representation is not None and
                    record.representation.codelist is not None and
                    not isinstance(record.representation.codelist, str)):
                data[record.id] = list(record.representation.codelist.
                                       items.keys())

    return data


def get_mandatory_attributes(dsd: DataStructureDefinition) -> list:
    data = []

    if (dsd.attribute_descriptor is None or
            dsd.attribute_descriptor.components is None):
        return []

    for record in dsd.attribute_descriptor.components.values():
        if (record.related_to is not None and
                record.assignment_status == "Mandatory"):
            data.append(record.id)

    return data


def validate_data(data: DataFrame, dsd: DataStructureDefinition):
    """
        Data validations stands for the next schema:

        .. list-table:: Data validations
            :widths: 20 80
            :header-rows: 1

            * - Code
              - Description
            * - SS01
              - Check if all dimensions of the DSD exist in the dataframe
            * - SS02
              - Check if the measure code exists in the dataframe
            * - SS03
              - Check if all mandatory attributes exist in the dataframe
            * - SS04
              - Check if an attribute/dimension value that is associated \
                to a Codelist is valid
            * - SS05
              - Check that all dimensions have values for every record of a \
                dataframe
            * - SS06
              - Check that all mandatory attributes have values for every \
                record of a dataset
            * - SS07
              - Check if two datapoints have the same values
            * - SS08
              - Check that the value inputted is compliant with Facets \
                for the referred Representation
            * - SS09
              - Check if the value is compliant with the desired Time Format
            * - SS10
              - Check if an attribute/dimension value associated to a \
                Cube Region Constraint is valid
            * - SS11
              - Check if each row is compliant with the Series Constraints
    """

    mandatory = get_mandatory_attributes(dsd)
    codelist_values = get_codelist_values(dsd)

    errors = []

    faceted, types = dsd._facet_type

    cubes, series_const = dsd._format_constraints

    process_measure_errors(dsd, data, errors, faceted, types)

    grouping_keys = []

    all_codes = dsd.dimension_codes + dsd.attribute_codes

    for k in dsd.dataset_attribute_codes:
        if k in all_codes:
            all_codes.remove(k)

    man_codes = dsd.dimension_codes + mandatory

    for k in all_codes:
        if k not in data.keys() or data[k].isnull().values.all():
            if k in mandatory:
                errors.append(
                    {'Code': 'SS03', 'ErrorLevel': 'CRITICAL',
                     'Component': f'{k}', 'Type': 'Attribute', 'Rows': None,
                     'Message': f'Missing {k}'})
            elif k in dsd.dimension_codes:
                errors.append(
                    {'Code': 'SS01', 'ErrorLevel': 'CRITICAL',
                     'Component': f'{k}', 'Type': 'Dimension', 'Rows': None,
                     'Message': f'Missing {k}'})
            continue
        else:
            process_errors_by_column(data, dsd, errors, k, man_codes,
                                     grouping_keys,
                                     faceted, types, codelist_values, cubes)

    if len(series_const) > 0:
        error_SS11(data, series_const, errors)

    if len(grouping_keys) > 0:
        duplicated = data[data.duplicated(subset=grouping_keys, keep=False)]
        if len(duplicated) > 0:
            duplicated.groupby(by=grouping_keys).apply(
                lambda x: create_error_SS07(x, errors, grouping_keys)
            )
        del duplicated

    return errors


# Time validations

# Monthly
regex_monthly = r'(19|[2-9][0-9])\d{2}(-(0[1-9]|1[012]))?'
match_monthly = re.compile(regex_monthly)

# Duration
regex_duration = r'(-?)P(?=.)((\d+)Y)?((\d+)M)?((\d+)D)?(T(?=.)((\d+)H)?((' \
                 r'\d+)M)?(\d*(\.\d+)?S)?)? '
match_duration = re.compile(regex_duration)

# semester, quarter and trimester
regex_specials = r'(19|[2-9][0-9])\d{2}-(A[1]|S[1-2]|Q[1-4]|T[' \
                 r'1-3]|M(0[1-9]|1[012])|W(5[0-3]|[1-4][0-9]|[' \
                 r'1-9])|D((00[1-9]|0[1-9][0-9])?|[12][0-9][0-9]|3[' \
                 r'0-5][0-9]|36[0-5]))'
match_specials = re.compile(regex_specials)

# Dictionaries

time_periods = ['ObservationalTimePeriod', 'ReportingTimePeriod',
                'GregorianTimePeriod', 'BasicTimePeriod', "StandardTimePeriod"]

gregorian_periods = {'GregorianDay': "%Y-%m-%d", 'GregorianYearMonth': "%Y-%m",
                     'GregorianYear': "%Y"}

reporting_periods = {'ReportingYear': "A[1]", 'ReportingSemester': "S[1-2]",
                     'ReportingTrimester': "T[1-3]",
                     'ReportingQuarter': "Q[1-4]",
                     "ReportingMonth": "M(0[1-9]|1[012])",
                     "ReportingWeek": "W(5[0-3]|[1-4][0-9]|[1-9])",
                     "ReportingDay": "D((00[1-9]|0[1-9][0-9])?"
                                     "|[12][0-9][0-9]|3[0-5][0-9]|36[0-5])"
                     }


def check_date(e, format_: str):
    try:
        res = datetime.strptime(e, format_)
        if not 1900 < res.year <= 9999:
            return False
    except ValueError:
        return False

    return True


def check_reporting(e, format_):
    # Matching semester, quarter and trimester
    match_reporting = re.compile(format_)

    try:
        res = match_reporting.fullmatch(e)
        if not res:
            return False
    except (TypeError, ValueError):
        return False
    return True


def validate_daily_iso(dt_str, type_):
    # Matching daily and iso format
    duration = ""
    control_changed = False
    if ((type_ == "ObservationalTimePeriod" or
         type_ == "StandardTimePeriod" or
         type_ == "BasicTimePeriod") and '/' in dt_str):
        duration = dt_str.split('/', maxsplit=1)[1]
        dt_str = dt_str.split('/', maxsplit=1)[0]

        # Matching duration
        try:
            res = match_duration.fullmatch(duration)
            if not res:
                dt_str += '/' + duration
                return False
        except (TypeError, ValueError):
            dt_str += '/' + duration
            return False
        control_changed = True
    try:
        res = match_specials.fullmatch(dt_str)
        if res:
            return True
    except (TypeError, ValueError):
        return False

    try:
        res = datetime.strptime(dt_str, '%Y-%m-%d')
        if res:
            return True
    except (TypeError, ValueError):
        return False

    try:
        res = datetime.fromisoformat(dt_str)
        if not 1900 < res.year <= 9999:
            if control_changed:
                dt_str += '/' + duration
            return False
    except (TypeError, ValueError):
        if control_changed:
            dt_str += '/' + duration
        return False
    if control_changed:
        dt_str += '/' + duration


def process_global_time_format(dt_str, type_):
    control_special = False
    for letter in ['A', 'Q', 'W', 'D']:
        if letter in dt_str:
            control_special = True

    if 'M' in dt_str and ':' not in dt_str:
        control_special = True

    if control_special:
        return process_special_time_format(dt_str)

    return process_common_time_format(dt_str, type_)


def process_common_time_format(dt_str, type_):
    # Matching year and monthly
    try:
        res = match_monthly.fullmatch(dt_str)
        if res:
            return True
    except (TypeError, ValueError):
        return False

    # Checks for datetime string in GregorianTimePeriod type
    if 'T' in dt_str and type_ == "GregorianTimePeriod":
        return False

    return validate_daily_iso(dt_str, type_)


def process_special_time_format(dt_str):
    try:
        res = match_specials.fullmatch(dt_str)
        return bool(res)
    except (TypeError, ValueError):
        return False


def time_period_valid(dt_str: str, type_: str):
    """Validates any time period"""

    if type_ in ["ObservationalTimePeriod", "StandardTimePeriod"]:
        return process_global_time_format(dt_str, type_)
    if type_ in ["GregorianTimePeriod", "BasicTimePeriod"]:
        return process_common_time_format(dt_str, type_)

    return process_special_time_format(dt_str)


def error_SS09(k, time_types, data_column, errors, role):
    if time_types[k] in time_periods:
        create_error_SS09(data_column, time_types[k], time_types[k], k,
                          role, errors, time_period_valid)
    elif time_types[k] in gregorian_periods:
        format_ = gregorian_periods[time_types[k]]
        create_error_SS09(data_column, format_, time_types[k], k,
                          role, errors, check_date)

    elif time_types[k] in reporting_periods:

        format_ = r'(19|[2-9][0-9])\d{2}-' + reporting_periods[time_types[k]]

        create_error_SS09(data_column, format_, time_types[k], k,
                          role, errors, time_period_valid)

    elif time_types[k].lower() == "datetime" or time_types[k] == "TimeRange":
        for e in data_column:
            invalid_date = False
            duration = ""
            control_changed = False
            if time_types[k] == "TimeRange" and '/' not in e:
                invalid_date = True
            elif time_types[k] == "TimeRange":
                duration = e.split('/', maxsplit=1)[1]
                e = e.split('/', maxsplit=1)[0]

                try:
                    res = match_duration.fullmatch(duration)
                    if not res:
                        invalid_date = True
                except (TypeError, ValueError):
                    invalid_date = True
                duration = '/' + duration
                control_changed = True

            try:
                res = datetime.fromisoformat(e)
                if not 1900 < res.year <= 9999:
                    invalid_date = True
            except (TypeError, ValueError):
                invalid_date = True

            if invalid_date:
                errors.append(
                    {'Code': 'SS09', 'ErrorLevel': "CRITICAL",
                     'Component': f'{k}',
                     'Type': f'{role}',
                     'Rows': None,
                     'Message': f'Value {e + duration} not compliant '
                                f'with type : {time_types[k]}'})
            if control_changed:
                e += duration


# Facets
def trunc_dec(x):
    return x.rstrip('0').rstrip('.') if '.' in x else x


def check_sequence(data_column, errors, key, type_, start, interval, end):
    data_column = np.sort(data_column)
    control = True
    if int(data_column[0]) < start:
        control = False
        values = data_column[data_column < start].tolist()
        if len(values) > 0:
            for v in values:
                errors.append(
                    {'Code': 'SS08', 'ErrorLevel': error_level_facets,
                     'Component': f'{key}', 'Type': f'{type_}',
                     'Rows': None,
                     'Message': f'Value {v} not compliant '
                                f'with startValue : {start}'})

    if end is not None and int(data_column[-1]) > end:
        control = False
        values = data_column[data_column > end].tolist()
        if len(values) > 0:
            for v in values:
                errors.append(
                    {'Code': 'SS08', 'ErrorLevel': error_level_facets,
                     'Component': f'{key}', 'Type': f'{type_}',
                     'Rows': None,
                     'Message': f'Value {v} not compliant '
                                f'with endValue : {end}'})

    if control:
        values = data_column[(data_column - start) % interval != 0]
        if len(values) > 0:
            for v in values:
                if end is not None:
                    errors.append(
                        {'Code': 'SS08', 'ErrorLevel': error_level_facets,
                         'Component': f'{key}',
                         'Type': f'{type_}',
                         'Rows': None,
                         'Message': f'Value {v} in {key} not compliant '
                                    f'with sequence : {start}-{end} '
                                    f'(interval: {interval})'})
                else:
                    errors.append(
                        {'Code': 'SS08', 'ErrorLevel': error_level_facets,
                         'Component': f'{key}',
                         'Type': f'{type_}',
                         'Rows': None,
                         'Message': f'Value {v} in {key} '
                                    f'not compliant with sequence : '
                                    f'{start}-infinite '
                                    f'(interval: {interval})'})


def check_num_facets(facets, data_column, key, type_):
    errors = []
    is_sequence = None
    start = None
    end = None
    interval = None
    for f in facets:
        values = []
        if f.facet_type == 'maxLength' or f.facet_type == 'minLength':
            temp = data_column.astype('str')
            temp = temp[np.isin(temp, ['nan', 'None'], invert=True)]
            format_temp = np.vectorize(trunc_dec)
            length_checker = np.vectorize(len)
            arr_len = length_checker(format_temp(temp))
            if f.facet_type == 'maxLength':
                max_ = int(f.facet_value)
                values = temp[arr_len > max_]
            else:
                min_ = int(f.facet_value)
                values = temp[arr_len < min_]

        elif f.facet_type == 'maxValue':
            max_ = int(f.facet_value)
            values = data_column[data_column > max_]
        elif f.facet_type == 'minValue':
            min_ = int(f.facet_value)
            values = data_column[data_column < min_]
        elif f.facet_type == 'isSequence':
            if f.facet_value.upper() == 'TRUE':
                is_sequence = True
        elif f.facet_type == 'startValue':
            start = int(f.facet_value)
        elif f.facet_type == 'endValue':
            end = int(f.facet_value)
        elif f.facet_type == 'interval':
            interval = int(f.facet_value)
        else:
            continue

        if len(values) > 0:
            for v in values:
                errors.append(
                    {'Code': 'SS08', 'ErrorLevel': error_level_facets,
                     'Component': f'{key}', 'Type': f'{type_}',
                     'Rows': None,
                     'Message': f'Value {v} not compliant with '
                                f'{f.facet_type} : {f.facet_value}'})

    if is_sequence is not None and start is not None and interval is not None:
        check_sequence(data_column, errors, key, type_, start, interval, end)

    return errors


def check_str_facets(facets, data_column, key, type_):
    data_column = data_column[
        np.isin(data_column, ['nan', 'None'], invert=True)].astype('str')

    error_level = 'WARNING'
    errors = []

    for f in facets:
        if f.facet_type == 'maxLength' or f.facet_type == 'maxValue':
            max_ = int(f.facet_value)
            if f.facet_type == 'maxLength':
                length_checker = np.vectorize(len)
                arr_len = length_checker(data_column)
                values = data_column[arr_len > max_]
            else:
                values = data_column[int(data_column) > max_]
        elif f.facet_type == 'minLength' or f.facet_type == 'minValue':
            min_ = int(f.facet_value)
            if f.facet_type == 'minLength':
                length_checker = np.vectorize(len)
                arr_len = length_checker(data_column)
                values = data_column[arr_len < min_]
            else:
                values = data_column[int(data_column) < min_]

        elif f.facet_type == 'pattern':
            r = re.compile(str(f.facet_value))
            vec = np.vectorize(lambda x: bool(not r.fullmatch(x)))
            values = data_column[vec(data_column)]
        else:
            continue

        if len(values) > 0:
            for v in values:
                errors.append({'Code': 'SS08', 'ErrorLevel': error_level,
                               'Component': f'{key}', 'Type': f'{type_}',
                               'Rows': None,
                               'Message': f'Value {v} not compliant with '
                                          f'{f.facet_type} : {f.facet_value}'})
    return errors


# Processing errors

def process_measure_errors(dsd, data, errors, faceted, types):
    mc = dsd.measure_code
    type_ = 'Measure'

    if mc not in data.keys() or data[mc].isnull().values.all():
        errors.append(
            {'Code': 'SS02', 'ErrorLevel': 'CRITICAL', 'Component': f'{mc}',
             'Type': f'{type_}', 'Rows': None,
             'Message': f'Missing {mc}'})
    elif data[mc].isnull().values.any():
        if 'OBS_STATUS' in data.keys():

            rows = data[
                (data[mc].isna()) & (data['OBS_STATUS'] != 'M')].to_dict(
                'records')
            if len(rows) > 0:
                errors.append({'Code': 'SS02', 'ErrorLevel': 'CRITICAL',
                               'Component': f'{mc}', 'Type': f'{type_}',
                               'Rows': rows.copy(),
                               'Message': f'Missing value in '
                                          f'{type_.lower()} {mc}'})
        else:
            rows = data[data[mc].isna()].to_dict('records')
            if len(rows) > 0:
                errors.append({'Code': 'SS02', 'ErrorLevel': 'CRITICAL',
                               'Component': f'{mc}', 'Type': f'{type_}',
                               'Rows': rows.copy(),
                               'Message': f'Missing value in '
                                          f'{type_.lower()} {mc}'})

    if mc in data.keys() and mc in faceted:
        facets = faceted[mc]
        try:
            data_column = data[mc].astype('float64')
            errors += check_num_facets(facets, data_column, mc, type_)
        except (TypeError, ValueError):
            data_column = data[mc].astype('str')
            errors += check_str_facets(facets, data_column, mc, type_)

        del data_column


def process_errors_by_column(data, dsd, errors, k, man_codes, grouping_keys,
                             faceted, types, codelist_values, cubes):
    is_numeric = False

    data_column = data[k].unique().astype('str')

    if 'TimePeriod' not in types[k]:
        try:
            data_column = data_column.astype('float64')
            is_numeric = True
        except (TypeError, ValueError):
            pass

    if k in dsd.attribute_codes:
        role = 'Attribute'
        code = 'SS06'
    else:
        grouping_keys.append(k)
        role = 'Dimension'
        code = 'SS05'

    if k in man_codes:
        control = False
        if is_numeric:
            if np.isnan(np.sum(data_column)):
                control = True
        else:
            if 'nan' in data_column:
                control = True
        if control:
            pos = data[(data[k] == 'nan') | (data[k] == '') |
                       (data[k].isnull())].index.values
            rows = data.iloc[pos, :].to_dict('records')
            errors.append({'Code': code, 'ErrorLevel': 'CRITICAL',
                           'Component': f'{k}', 'Type': f'{role}',
                           'Rows': rows.copy(),
                           'Message': f'Missing value in '
                                      f'{role.lower()} {k}'})

    if k in types:
        error_SS09(k, types, data_column, errors, role)

    if k in faceted:
        facets = faceted[k]
        if is_numeric:
            errors += check_num_facets(facets, data_column, k, role)
        else:
            errors += check_str_facets(facets, data_column, k, role)

    if is_numeric:
        data_column = data_column.astype('str')
        format_temp = np.vectorize(trunc_dec)
        data_column = format_temp(data_column)

    if k in cubes.keys():
        code = 'SS10'
        values = data_column[
            np.isin(data_column, list(cubes[k]), invert=True)]
        if len(values) > 0:
            create_error_SS10_SS04(values, code, role, k, errors)

    elif k in codelist_values.keys():
        code = 'SS04'
        values = data_column[
            np.isin(data_column, codelist_values[k], invert=True)]
        if len(values) > 0:
            create_error_SS10_SS04(values, code, role, k, errors)


def create_error_SS10_SS04(values, code, role, k, errors):
    values = values[
        np.isin(values, ['nan', 'None', np.nan], invert=True)]
    for v in values:
        errors.append({'Code': code, 'ErrorLevel': 'CRITICAL',
                       'Component': f'{k}', 'Type': f'{role}',
                       'Rows': None,
                       'Message': f'Wrong value {v} for '
                                  f'{role.lower()} {k}'})


def format_row(row):
    string = ''
    for k, v in row.items():
        string += f' ( {str(k)} : {str(v) if str(v) != "nan" else ""} ) '
    return string


def create_error_SS07(x, errors, grouping_keys):
    errors.append({'Code': 'SS07',
                   'ErrorLevel': 'WARNING',
                   'Component': 'Duplicated',
                   'Type': 'Datapoint',
                   'Rows': x.to_dict(orient="records").copy(),
                   'Message': f'Duplicated datapoint '
                              f'{format_row(x.loc[0, grouping_keys])}'
                   })


def create_error_SS09(data_column, format_, time_type, comp, role, errors,
                      func):
    for e in data_column:
        if not func(e, format_):
            errors.append(
                {'Code': 'SS09', 'ErrorLevel': "CRITICAL",
                 'Component': f'{comp}',
                 'Type': f'{role}',
                 'Rows': None,
                 'Message': f'Value {e} not compliant with '
                            f'type : {time_type}'})


def error_SS11(data, series_const, errors):
    lookup = pd.DataFrame(series_const).drop_duplicates() \
        .reset_index(drop=True)
    all_columns = lookup.columns.tolist()

    result = all(elem in data.columns.tolist() for elem in all_columns)

    if result:
        columns = lookup.columns[lookup.isna().any()].tolist()

        lookup['membership_series_const'] = True

        dict_wild = {}

        for e in columns:
            dict_wild[e] = lookup[lookup[e].isna()]
            dict_wild[e].pop(e)

        res = data[all_columns].merge(lookup, how="left")

        for k in dict_wild:
            res.update(data[all_columns].merge(dict_wild[k], how="left"),
                       overwrite=False)

        indexes = res[res['membership_series_const'].isna()].index.tolist()

        del res
        del dict_wild
        del lookup
        del all_columns

        if len(indexes) > 0:
            rows = data.loc[indexes, :].to_dict('records')
            errors.append({'Code': 'SS11',
                           'ErrorLevel': 'WARNING',
                           'Component': 'Series',
                           'Type': 'Constraint',
                           'Rows': rows.copy(),
                           'Message': 'Found disallowed rows'
                           })
            del rows
    else:
        del lookup
        del all_columns



from sdmxthon.model.base import IdentifiableArtefact
from sdmxthon.model.itemScheme import Concept
from sdmxthon.model.representation import Representation
from sdmxthon.model.utils import generic_setter, int_setter
from sdmxthon.utils.handlers import add_indent, export_intern_data, \
    split_unique_id
from sdmxthon.utils.mappings import structureAbbr


class Component(IdentifiableArtefact):
    """ A component is an abstract super class used to define qualitative and
        quantitative data and metadata items that belong to a Component List
        and hence a Structure. Component is refined through its sub-classes."""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, local_representation: Representation = None,
                 concept_identity: Concept = None):
        super(Component, self).__init__(id_=id_, uri=uri, urn=urn,
                                        annotations=annotations)

        self.local_representation = local_representation
        self.concept_identity = concept_identity

    def __eq__(self, other):
        if isinstance(other, Component):
            return (super(Component, self).__eq__(other) and
                    self.local_representation == other.local_representation and
                    self._concept_identity == other._concept_identity)

        return False

    def __str__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __unicode__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} - {self.id}>'

    @property
    def local_representation(self):
        """Association to the Representation of the Component if this is
        different from the coreRepresentation of the Concept which the
        Component uses (ConceptUsage) """
        return self._local_representation

    @local_representation.setter
    def local_representation(self, value: Representation):
        self._local_representation = generic_setter(value, Representation)

    @property
    def concept_identity(self):
        """Association to a Concept in a Concept Scheme that
            identifies and defines the semantic of the Component"""
        return self._concept_identity

    @concept_identity.setter
    def concept_identity(self, value: Concept):
        self._concept_identity = generic_setter(value, Concept)

    @property
    def representation(self):
        """Extracts the representation from a Component"""
        if self.local_representation is None and self.concept_identity is None:
            return None

        if self.local_representation is not None:
            return self.local_representation

        if isinstance(self.concept_identity, Concept) and \
                self.concept_identity.core_representation is not None:
            return self.concept_identity.core_representation

        return None

    def _parse_XML(self, indent, head):

        if isinstance(self, TimeDimension):
            head = 'TimeDimension'

        head = f'{structureAbbr}:' + head

        prettyprint = indent != ''

        data = super(Component, self)._to_XML(prettyprint)
        indent = add_indent(indent)
        if isinstance(self, Dimension):
            outfile = f'{indent}<{head}{data["Attributes"]} ' \
                      f'position="{self._position}">'
        elif isinstance(self, Attribute):
            outfile = f'{indent}<{head}{data["Attributes"]} ' \
                      f'assignmentStatus="{self._usageStatus}">'
        else:
            outfile = f'{indent}<{head}{data["Attributes"]}>'
        outfile += export_intern_data(data, indent)

        indent_child = add_indent(indent)

        if self.concept_identity is not None:
            indent_ref = add_indent(indent_child)
            outfile += f'{indent_child}<{structureAbbr}:ConceptIdentity>'

            if isinstance(self.concept_identity, dict):
                agencyID, id_, version = \
                    split_unique_id(self.concept_identity['CS'])

                outfile += f'{indent_ref}<Ref maintainableParentID="{id_}" ' \
                           f'package="conceptscheme" ' \
                           f'maintainableParentVersion="{version}" ' \
                           f'agencyID="{agencyID}" ' \
                           f'id="{self.concept_identity["CON"]}" ' \
                           f'class="Concept"/>'

            else:
                outfile += f'{indent_ref}<Ref ' \
                           f'maintainableParentID=' \
                           f'"{self.concept_identity.scheme.id}" ' \
                           f'package="conceptscheme" ' \
                           f'maintainableParentVersion=' \
                           f'"{self.concept_identity.scheme.version}" ' \
                           f'agencyID=' \
                           f'"{self.concept_identity.scheme.agencyID}" ' \
                           f'id="{self.concept_identity.id}" ' \
                           f'class="Concept"/>'

            outfile += f'{indent_child}</{structureAbbr}:ConceptIdentity>'
        indent_enum = add_indent(indent_child)
        indent_ref = add_indent(indent_enum)
        if self.local_representation is not None:
            outfile += f'{indent_child}<{structureAbbr}:LocalRepresentation>'
            if self.local_representation.codelist is not None:
                label_format = 'EnumerationFormat'
                outfile += f'{indent_enum}<{structureAbbr}:Enumeration>'
                if isinstance(self.local_representation.codelist, str):
                    agencyID, id_, version = \
                        split_unique_id(self.local_representation.codelist)

                    outfile += f'{indent_ref}<Ref package="codelist" ' \
                               f'agencyID="{agencyID}" ' \
                               f'id="{id_}" ' \
                               f'version="{version}" class="Codelist"/>'
                else:
                    agencyID = self.local_representation.codelist.agencyID
                    id_ = self.local_representation.codelist.id
                    version = self.local_representation.codelist.version
                    outfile += f'{indent_ref}<Ref package="codelist" ' \
                               f'agencyID="{agencyID}" id="{id_}" ' \
                               f'version="{version}" class="Codelist"/>'

                outfile += f'{indent_enum}</{structureAbbr}:Enumeration>'
            else:
                label_format = 'TextFormat'

            format_attributes = ' '

            if self.local_representation.type_ is not None:
                format_attributes = f' textType=' \
                                    f'"{self.local_representation.type_}"'

            if self.local_representation.facets is not None:
                for e in self.local_representation.facets:
                    format_attributes += f' {e.facet_type}="{e.facet_value}"'

            outfile += f'{indent_enum}<{structureAbbr}:' \
                       f'{label_format}{format_attributes}/>'

            outfile += f'{indent_child}</{structureAbbr}:LocalRepresentation>'

        if head == f'{structureAbbr}:Attribute':
            outfile += f'{indent_child}<{structureAbbr}:' \
                       f'AttributeRelationship>'

            if isinstance(self.related_to, dict):
                if 'id' in self.related_to.keys():

                    if isinstance(self.related_to['id'], list):
                        for k in self.related_to['id']:
                            outfile += f'{indent_enum}<{structureAbbr}:' \
                                       f'{self.related_to["type"]}>'
                            outfile += f'{indent_ref}<Ref id="{k}"/>'
                            outfile += f'{indent_enum}</{structureAbbr}:' \
                                       f'{self.related_to["type"]}>'
                    else:
                        outfile += f'{indent_enum}<{structureAbbr}:' \
                                   f'{self.related_to["type"]}>'
                        outfile += f'{indent_ref}<Ref ' \
                                   f'id="{self.related_to["id"]}"/>'
                        outfile += f'{indent_enum}</{structureAbbr}:' \
                                   f'{self.related_to["type"]}>'
                else:
                    for k in self.related_to:
                        outfile += f'{indent_enum}<{structureAbbr}' \
                                   f':Dimension>'
                        outfile += f'{indent_ref}<Ref id="{k}"/>'
                        outfile += f'{indent_enum}</{structureAbbr}' \
                                   f':Dimension>'
            elif isinstance(self.related_to, Dimension):
                outfile += f'{indent_enum}<{structureAbbr}:Dimension>'
                outfile += f'{indent_ref}<Ref id="{self.related_to.id}"/>'
                outfile += f'{indent_enum}</{structureAbbr}:Dimension>'
            elif isinstance(self.related_to, PrimaryMeasure):
                outfile += f'{indent_enum}<{structureAbbr}' \
                           f':PrimaryMeasure>'
                outfile += f'{indent_ref}<Ref id="{self.related_to.id}"/>'
                outfile += f'{indent_enum}</{structureAbbr}' \
                           f':PrimaryMeasure>'
            else:
                outfile += f'{indent_enum}<{structureAbbr}:None/>'

            outfile += f'{indent_child}</{structureAbbr}' \
                       f':AttributeRelationship>'

        outfile += f'{indent}</{head}>'
        return outfile


class Dimension(Component):
    """ A metadata concept used (most probably together with other metadata
        concepts) to classify a statistical series, e.g. a statistical concept
        indicating a certain economic activity or a geographical
        reference area."""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None, local_representation: Representation = None,
                 concept_identity: Concept = None,
                 position: int = None):
        super(Dimension, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     local_representation=local_representation,
                     concept_identity=concept_identity)

        self.position = position

    def __eq__(self, other):
        if isinstance(other, Dimension):
            return super(Dimension, self).__eq__(
                other) and self._position == other._position
        return False

    @property
    def position(self):
        """Position of the Dimension in the DimensionList"""
        return self._position

    @position.setter
    def position(self, value):
        self._position = int_setter(value)


class TimeDimension(Dimension):
    """A metadata concept that identifies the component in the key structure
       that has the role of “time”.
    """

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 local_representation: Representation = None,
                 concept_identity: Concept = None,
                 position: int = None):
        super(TimeDimension, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     local_representation=local_representation,
                     concept_identity=concept_identity,
                     position=position)

    def __eq__(self, other):
        if isinstance(other, TimeDimension):
            return super(TimeDimension, self).__eq__(other)

        return False


class Attribute(Component):
    """A characteristic of an object or entity."""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 local_representation: Representation = None,
                 concept_identity: Concept = None,
                 assignmentStatus: str = None,
                 relatedTo=None):
        super(Attribute, self) \
            .__init__(id_=id_, uri=uri, urn=urn,
                      annotations=annotations,
                      local_representation=local_representation,
                      concept_identity=concept_identity)

        self.assignment_status = assignmentStatus
        self.related_to = relatedTo

    def __eq__(self, other):
        if isinstance(other, Attribute):
            return (super(Attribute, self).__eq__(other) and
                    self._usageStatus == other._usageStatus and
                    self._relatedTo == other._relatedTo)

        return False

    @property
    def assignment_status(self):
        """Defines the usage status of the Attribute
        (Mandatory, Conditional)"""
        return self._usageStatus

    @property
    def related_to(self):
        """Association to a AttributeRelationship."""
        return self._relatedTo

    @assignment_status.setter
    def assignment_status(self, value):
        if value in ["Mandatory", "Conditional"] or value is None:
            self._usageStatus = value
        else:
            raise ValueError(
                "The value for usageStatus has to be 'Mandatory' "
                "or 'Conditional'")

    @related_to.setter
    def related_to(self, value):
        if value is None:
            self._relatedTo = "NoSpecifiedRelationship"
        else:
            self._relatedTo = value


class MeasureDimension(Dimension):
    """ A statistical concept that identifies the component in the key
        structure that has an enumerated list of measures. This dimension has,
        as its representation the Concept Scheme that enumerates the
        measure concepts.
    """

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 local_representation: Representation = None,
                 concept_identity: Concept = None,
                 position: int = None):
        super(MeasureDimension, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     local_representation=local_representation,
                     concept_identity=concept_identity,
                     position=position)

    def __eq__(self, other):
        if isinstance(other, MeasureDimension):
            return super(MeasureDimension, self).__eq__(other)

        return False


class PrimaryMeasure(Component):
    """The metadata concept that is the phenomenon to be measured in a
    data set. In a data set the instance of the measure is often called
    the observation."""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 local_representation: Representation = None,
                 concept_identity: Concept = None):
        super(PrimaryMeasure, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     concept_identity=concept_identity,
                     local_representation=local_representation)

    def __eq__(self, other):
        if isinstance(other, PrimaryMeasure):
            return super(PrimaryMeasure, self).__eq__(other)

        return False



from datetime import datetime, date, timedelta

from sdmxthon.model.utils import generic_setter


class Period:
    def __init__(self, isInclusive: bool, period: datetime):
        self.is_inclusive = isInclusive
        self.period = period

    @property
    def is_inclusive(self):
        return self._isInclusive

    @is_inclusive.setter
    def is_inclusive(self, value):
        self._isInclusive = generic_setter(value, bool)

    @property
    def period(self):
        return self._period

    @period.setter
    def period(self, value):
        self._period = generic_setter(value, datetime)


class ReferencePeriod:
    def __init__(self, start: date, end: date):
        self.start_date = start
        self.end_date = end

    @property
    def start_date(self):
        return self._start

    @start_date.setter
    def start_date(self, value):
        self._start = generic_setter(value, date)

    @property
    def end_date(self):
        return self._end

    @end_date.setter
    def end_date(self, value):
        self._end = generic_setter(value, date)


class ReleaseCalendar:
    """
    javax.xml.datatype.Duration is substituted in this method with
    datetime.timedelta, as they have a similar meaning for the parsing
    methods used
    """

    def __init__(self, periodicity: timedelta, offset: timedelta,
                 tolerance: timedelta):
        self.periodicity = periodicity
        self.offset = offset
        self.tolerance = tolerance

    @property
    def periodicity(self):
        return self._periodicity

    @periodicity.setter
    def periodicity(self, value):
        self._periodicity = generic_setter(value, timedelta)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = generic_setter(value, timedelta)

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value):
        self._tolerance = generic_setter(value, timedelta)



"""
    Component file holds the classes for the Component and its derivatives
"""

import json
from datetime import datetime
from typing import List

from sdmxthon.model.base import MaintainableArtefact, \
    InternationalString
from sdmxthon.model.component import Component
from sdmxthon.model.descriptors import ComponentList, DimensionDescriptor, \
    AttributeDescriptor, MeasureDescriptor, GroupDimensionDescriptor
from sdmxthon.model.extras import ReferencePeriod, ReleaseCalendar
from sdmxthon.model.utils import generic_setter, ConstraintRoleType, \
    bool_setter
from sdmxthon.utils.handlers import export_intern_data, add_indent, \
    split_unique_id
from sdmxthon.utils.mappings import structureAbbr, Data_Types_VTL, commonAbbr


class MemberSelection(object):

    def __init__(self, is_included: bool = False,
                 values_for: Component = None, sel_value: list = None):
        self.is_included = is_included
        self.values_for = values_for
        self.sel_value = []
        if sel_value is not None:
            self.sel_value = sel_value

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of MemberSelection"""
        return MemberSelection(*args_, **kwargs_)

    @property
    def sel_value(self) -> list:
        return self._selValue

    @sel_value.setter
    def sel_value(self, value):
        self._selValue = generic_setter(value, list)

    @property
    def values_for(self):
        return self._values_for

    @values_for.setter
    def values_for(self, value):
        self._values_for = generic_setter(value, str)

    @property
    def is_included(self):
        return self._is_included

    @is_included.setter
    def is_included(self, value):
        self._is_included = bool_setter(value)


class CubeRegion(object):

    def __init__(self, is_included: bool = False,
                 member: List[MemberSelection] = None):
        self.is_included = is_included

        self.member = member

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of CubeRegion"""
        return CubeRegion(*args_, **kwargs_)

    @property
    def member(self):
        return self._member

    @member.setter
    def member(self, value):
        self._member = generic_setter(value, list)

    @property
    def is_included(self):
        return self._is_included

    @is_included.setter
    def is_included(self, value):
        self._is_included = bool_setter(value)


class MetadataTargetRegion:

    def __init__(self, comp_list: ComponentList, isIncluded: bool = False,
                 member: MemberSelection = None):
        self.is_included = isIncluded

        self.comp_list = comp_list

        self.member = member

    @property
    def member(self):
        return self._member

    @member.setter
    def member(self, value):
        self._member = generic_setter(value, MemberSelection)

    @property
    def comp_list(self):
        return self._compList

    @comp_list.setter
    def comp_list(self, value):
        self._compList = generic_setter(value, ComponentList)

    @property
    def is_included(self):
        return self._isIncluded

    @is_included.setter
    def is_included(self, value):
        self._isIncluded = generic_setter(value, bool)


class DataKeySet(object):

    def __init__(self, keys: list = None, isIncluded: bool = None):
        self.keys = []

        if keys is not None:
            self.keys = keys

        self._isIncluded = isIncluded

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of DataKeySet"""
        return DataKeySet(*args_, **kwargs_)

    @property
    def keys(self):
        return self._keys

    @keys.setter
    def keys(self, value):
        self._keys = generic_setter(value, list)

    @property
    def is_included(self):
        return self._isIncluded

    @is_included.setter
    def is_included(self, value):
        self._isIncluded = bool_setter(value)


class Constraint(MaintainableArtefact):
    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None,
                 maintainer=None,
                 validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None,
                 isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None,
                 dataContentRegion: List[CubeRegion] = None,
                 dataKeySet: List[DataKeySet] = None,
                 metadataContentRegion: List[MetadataTargetRegion] = None,
                 availableDates: List[ReferencePeriod] = None,
                 calendar: List[ReleaseCalendar] = None):
        if annotations is None:
            annotations = []
        super(Constraint, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name, description=description,
                     version=version, validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     maintainer=maintainer,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl)
        self.metadata_content_region = []
        self.data_content_region = []
        self.available_dates = []
        self.calendar = []

        if dataContentRegion is not None:
            self.data_content_region = dataContentRegion

        if metadataContentRegion is not None:
            self.metadata_content_region = metadataContentRegion
        if availableDates is not None:
            self.available_dates = availableDates
        if calendar is not None:
            self.calendar = calendar

        self.data_content_keys = dataKeySet

        self._ref_attach = None
        self._type_attach = None

    @property
    def data_content_region(self) -> list:
        return self._data_content_region

    @data_content_region.setter
    def data_content_region(self, value):
        self._data_content_region = generic_setter(value, list)

    @property
    def data_content_keys(self):
        return self._data_content_keys

    @data_content_keys.setter
    def data_content_keys(self, value):
        self._data_content_keys = generic_setter(value, list)

    @property
    def metadata_content_region(self):
        return self._metadata_content_region

    @metadata_content_region.setter
    def metadata_content_region(self, value):
        self._metadata_content_region = value

    @property
    def available_dates(self):
        return self._available_dates

    @available_dates.setter
    def available_dates(self, value):
        self._available_dates = generic_setter(value, list)

    @property
    def calendar(self):
        return self._calendar

    @calendar.setter
    def calendar(self, value):
        self._calendar = generic_setter(value, list)

    @property
    def ref_attach(self):
        return self._ref_attach

    @property
    def type_attach(self):
        return self._type_attach


class KeySetType(object):
    def __init__(self):
        self._key = {}

    @property
    def key(self):
        return self._key


class ContentConstraint(Constraint):
    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None,
                 maintainer=None,
                 validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None,
                 dataContentRegion: List[CubeRegion] = None,
                 dataKeySet: List[DataKeySet] = None,
                 metadataContentRegion: List[MetadataTargetRegion] = None,
                 availableDates: List[ReferencePeriod] = None,
                 calendar: List[ReleaseCalendar] = None,
                 role: str = None):
        if annotations is None:
            annotations = []
        super(ContentConstraint, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name,
                     description=description,
                     version=version,
                     maintainer=maintainer,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     dataContentRegion=dataContentRegion,
                     dataKeySet=dataKeySet,
                     metadataContentRegion=metadataContentRegion,
                     availableDates=availableDates,
                     calendar=calendar)
        if role is not None and role not in ConstraintRoleType:
            raise ValueError('ConstraintRole must be either '
                             '"allowableContent" or "actualContent"')
        self._role = role

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, value):
        if value not in ConstraintRoleType:
            raise ValueError('ConstraintRole must be either '
                             '"Allowed" or "Actual"')
        self._role = value

    def _parse_XML(self, indent, label):
        prettyprint = indent != ''

        indent = add_indent(indent)

        data = super(Constraint, self)._to_XML(prettyprint)

        outfile = ''

        attributes = data.get('Attributes') or None

        if attributes is not None:
            outfile += f'{indent}<{label}{attributes} type="{self.role}">'
        else:
            outfile += f'{indent}<{label}>'

        outfile += export_intern_data(data, indent)

        indent_child = add_indent(indent)
        indent_child_2 = add_indent(indent_child)
        indent_ref = add_indent(indent_child_2)
        indent_value = add_indent(indent_ref)

        outfile += f'{indent_child}<{structureAbbr}:ConstraintAttachment>'
        outfile += f'{indent_child_2}<{structureAbbr}:{self.type_attach}>'

        agencyID, id_, version = split_unique_id(self.ref_attach)

        outfile += f'{indent_ref}<Ref id="{id_}" ' \
                   f'version="{version}" ' \
                   f'agencyID="{agencyID}" ' \
                   f'package="datastructure" class="{self.type_attach}"/>'

        outfile += f'{indent_child_2}</{structureAbbr}:{self.type_attach}>'
        outfile += f'{indent_child}</{structureAbbr}:ConstraintAttachment>'

        if self.data_content_keys is not None:
            for i in self.data_content_keys:
                outfile += f'{indent_child}<{structureAbbr}:DataKeySet ' \
                           f'isIncluded="' \
                           f'{str(i.is_included).lower()}">'

                for e in i.keys:
                    outfile += f'{indent_child_2}<{structureAbbr}:Key>'

                    for k, v in e.items():
                        outfile += f'{indent_ref}<{commonAbbr}:KeyValue ' \
                                   f'id="{k}">'
                        outfile += f'{indent_value}<{commonAbbr}:Value>{v}' \
                                   f'</{commonAbbr}:Value>'
                        outfile += f'{indent_ref}</{commonAbbr}:KeyValue>'

                    outfile += f'{indent_child_2}</{structureAbbr}:Key>'

                outfile += f'{indent_child}</{structureAbbr}:DataKeySet>'

        if self.data_content_region is not None:

            for e in self.data_content_region:
                outfile += f'{indent_child}<{structureAbbr}:CubeRegion ' \
                           f'isIncluded="{str(e.is_included).lower()}">'
                for m in e.member:
                    outfile += f'{indent_child_2}<{commonAbbr}:KeyValue ' \
                               f'id="{m.values_for}">'
                    for j in m.sel_value:
                        outfile += f'{indent_ref}<{commonAbbr}:Value>{j}' \
                                   f'</{commonAbbr}:Value>'
                    outfile += f'{indent_child_2}</{commonAbbr}:KeyValue>'
                outfile += f'{indent_child}</{structureAbbr}:CubeRegion>'

        outfile += f'{indent}</{label}>'

        return outfile


def add_type_facets(key, value, type_, facets):
    if value.representation is not None:
        if value.representation.type_ is not None:
            type_[key] = value.representation.type_
        else:
            type_[key] = 'String'

        if len(value.representation.facets) > 0:
            facets[key] = value.representation.facets
    else:
        type_[key] = 'String'


class DataStructureDefinition(MaintainableArtefact):
    """A collection of metadata concepts, their structure and usage when used
    to collect or disseminate data."""

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 dimension_list: DimensionDescriptor = None,
                 measure_list: MeasureDescriptor = None,
                 attribute_list: AttributeDescriptor = None,
                 group_dimension_descriptor: GroupDimensionDescriptor = None,
                 constraint: list = None):

        super(DataStructureDefinition, self). \
            __init__(id_=id_, uri=uri,
                     urn=urn,
                     annotations=annotations,
                     name=name,
                     description=description,
                     version=version,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer)

        self.dimension_descriptor = dimension_list
        self.measure_descriptor = measure_list
        self.attribute_descriptor = attribute_list
        self.group_dimension_descriptor = group_dimension_descriptor
        self._constraints = constraint

    def __eq__(self, other):
        if isinstance(other, DataStructureDefinition):
            return (super.__eq__(self, other) and
                    self._dimensionDescriptor == other._dimensionDescriptor and
                    self._attributeDescriptor == other._attributeDescriptor and
                    self._measureDescriptor == other._measureDescriptor and
                    self._groupDimensionDescriptor ==
                    other._groupDimensionDescriptor)

        return False

    def __str__(self):
        return '<DataStructureDefinition  - %s:%s(%s)>' % (
            self.agencyID, self.id, self.version)

    def __unicode__(self):
        return u'<DataStructureDefinition  - %s:%s(%s)>' % (
            self.agencyID, self.id, self.version)

    def __repr__(self):
        return '<DataStructureDefinition  - %s:%s(%s)>' % (
            self.agencyID, self.id, self.version)

    @property
    def dimension_descriptor(self):
        """An ordered set of metadata concepts that, combined, classify a
        statistical series, and whose values, when combined (the key)
        in an instance such as a data set, uniquely identify a
        specific observation"""
        return self._dimensionDescriptor

    @property
    def measure_descriptor(self):
        """A metadata concept that defines the measure of a
        Data Structure Definition"""
        return self._measureDescriptor

    @property
    def attribute_descriptor(self):
        """A set metadata concepts that define the attributes of a Data
        Structure Definition. """
        return self._attributeDescriptor

    @property
    def group_dimension_descriptor(self):
        """A set metadata concepts that define a partial key derived
           from the Dimension Descriptor in a Data Structure Definition."""
        return self._groupDimensionDescriptor

    @property
    def dimension_codes(self):
        """Keys of the dimensionDescriptor components"""
        return [k for k in self.dimension_descriptor.components]

    @property
    def attribute_codes(self):
        """Keys of the attributeDescriptor components"""
        if self.attribute_descriptor is not None:
            return [k for k in self.attribute_descriptor.components]

        return []

    @property
    def dataset_attribute_codes(self):
        """Attributes with no specified relationship"""
        result = []
        if self.attribute_descriptor is not None:
            for k in self.attribute_descriptor.components:
                if self.attribute_descriptor[k].related_to == \
                        "NoSpecifiedRelationship":
                    result.append(k)
        return result

    @property
    def content(self):
        """Extracts the content of a DSD as a dict"""

        result = {'dimensions': self.dimension_descriptor.components,
                  'measure': self.measure_descriptor.components[
                      self.measure_code]
                  }

        if self.attribute_descriptor is not None:
            result['attributes'] = self.attribute_descriptor.components

        if self.group_dimension_descriptor is not None:
            result['groups'] = self.group_dimension_descriptor

        return result

    @property
    def _facet_type(self):
        """Returns any component that has facets"""
        facets = {}
        type_ = {}
        for k, v in self.dimension_descriptor.components.items():
            add_type_facets(k, v, type_, facets)

        if self.attribute_descriptor is not None:
            for k, v in self.attribute_descriptor.components.items():
                add_type_facets(k, v, type_, facets)

        if self.measure_descriptor is not None:
            for k, v in self.measure_descriptor.components.items():
                add_type_facets(k, v, type_, facets)

        return facets, type_

    @property
    def _format_constraints(self):
        """Returns the constraints in a formatted way for validation"""
        cubes = {}
        series = []

        if self.constraints is not None:
            for c in self.constraints:
                if len(c.data_content_region) > 0:
                    for e in c.data_content_region:
                        for m in e.member:
                            if m.values_for not in cubes.keys():
                                cubes[m.values_for] = set(
                                    m.sel_value)
                            else:
                                cubes[m.values_for].update(
                                    m.sel_value)
                if (c.data_content_keys is not None and
                        len(c.data_content_keys) > 0):
                    for e in c.data_content_keys:
                        if c.role is not None and c.role == "Allowed":
                            series += e.keys

        return cubes, series

    @property
    def measure_code(self):
        """Key of the MeasureDescriptor component (PrimaryMeasure)"""
        return list(self.measure_descriptor.components.keys())[0]

    @dimension_descriptor.setter
    def dimension_descriptor(self, value):
        self._dimensionDescriptor = generic_setter(value, DimensionDescriptor)

    @measure_descriptor.setter
    def measure_descriptor(self, value):
        self._measureDescriptor = generic_setter(value, MeasureDescriptor)

    @attribute_descriptor.setter
    def attribute_descriptor(self, value):
        self._attributeDescriptor = generic_setter(value, AttributeDescriptor)

    @group_dimension_descriptor.setter
    def group_dimension_descriptor(self, value):
        self._groupDimensionDescriptor = \
            generic_setter(value, GroupDimensionDescriptor)

    @property
    def constraints(self):
        return self._constraints

    def add_constraint(self, value: ContentConstraint):
        if self._constraints is None:
            self._constraints = []
        self._constraints.append(value)

    def to_vtl_json(self, path: str = None):
        """Formats the DataStructureDefinition as a VTL DataStructure"""
        dataset_name = self.id
        components = []
        for c in self.dimension_descriptor.components.values():

            type_ = "String"

            if (c.representation is not None and
                    c.representation.type_ is not None):
                type_ = c.representation.type_

            component = {"name": c.id, "role": "Identifier",
                         "type": Data_Types_VTL[type_], "isNull": False}

            components.append(component)
        if self.attribute_descriptor is not None:
            for c in self.attribute_descriptor.components.values():
                type_ = "String"

                if (c.representation is not None and
                        c.representation.type_ is not None):
                    type_ = c.representation.type_

                component = {"name": c.id, "role": "Attribute",
                             "type": Data_Types_VTL[type_], "isNull": True}

                components.append(component)
        for c in self.measure_descriptor.components.values():
            type_ = "String"

            if (c.representation is not None and
                    c.representation.type_ is not None):
                type_ = c.representation.type_

            component = {"name": c.id, "role": "Measure",
                         "type": Data_Types_VTL[type_], "isNull": True}

            components.append(component)

        result = {
            "DataSet": {"name": dataset_name, "DataStructure": components}}
        if path is not None:
            with open(path, 'w') as fp:
                fp.write(json.dumps(result))
        else:
            return result

    def _parse_XML(self, indent, label):
        prettyprint = indent != ''

        indent = add_indent(indent)

        data = super(DataStructureDefinition, self)._to_XML(prettyprint)

        outfile = ''

        attributes = data.get('Attributes') or None

        if attributes is not None:
            outfile += f'{indent}<{label}{attributes}>'
        else:
            outfile += f'{indent}<{label}>'

        outfile += export_intern_data(data, indent)

        indent_child = add_indent(indent)

        outfile += f'{indent_child}<{structureAbbr}:DataStructureComponents>'

        if self.dimension_descriptor is not None:
            outfile += self.dimension_descriptor. \
                _parse_XML(indent_child, f'{structureAbbr}:DimensionList')

        if self.attribute_descriptor is not None:
            outfile += self.attribute_descriptor. \
                _parse_XML(indent_child, f'{structureAbbr}:AttributeList')

        if self.measure_descriptor is not None:
            outfile += self.measure_descriptor. \
                _parse_XML(indent_child, f'{structureAbbr}:MeasureList')

        outfile += f'{indent_child}</{structureAbbr}:DataStructureComponents>'

        outfile += f'{indent}</{label}>'

        return outfile


class DataFlowDefinition(MaintainableArtefact):
    """Abstract concept (i.e. the structure without any data) of a flow of data
       that providers will provide for different reference periods.
    """

    def __init__(self, id_: str = None, uri: str = None, urn: str = None,
                 annotations=None,
                 name: InternationalString = None,
                 description: InternationalString = None,
                 version: str = None, validFrom: datetime = None,
                 validTo: datetime = None,
                 isFinal: bool = None, isExternalReference: bool = None,
                 serviceUrl: str = None,
                 structureUrl: str = None, maintainer=None,
                 structure: DataStructureDefinition = None,
                 constraints: list = None):
        super(DataFlowDefinition, self). \
            __init__(id_=id_, uri=uri, urn=urn,
                     annotations=annotations,
                     name=name,
                     description=description,
                     version=version,
                     validFrom=validFrom,
                     validTo=validTo,
                     isFinal=isFinal,
                     isExternalReference=isExternalReference,
                     serviceUrl=serviceUrl,
                     structureUrl=structureUrl,
                     maintainer=maintainer)
        self.structure = structure
        self._constraints = constraints

    def __eq__(self, other):
        if isinstance(other, DataFlowDefinition):
            return (super(DataFlowDefinition, self).__eq__(other) and
                    self._structure == other._structure)

        return False

    def __str__(self):
        return f'<DataFlowDefinition - {self.unique_id}>'

    def __unicode__(self):
        return f'<DataFlowDefinition - {self.unique_id}>'

    def __repr__(self):
        return f'<DataFlowDefinition - {self.unique_id}>'

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of DataFlowDefinition"""
        return DataFlowDefinition(*args_, **kwargs_)

    @property
    def structure(self):
        """Associates a DataflowDefinition to the DataStructureDefinition."""
        return self._structure

    @structure.setter
    def structure(self, value):
        self._structure = generic_setter(value, DataStructureDefinition)

    def add_constraint(self, value: ContentConstraint):
        if self._constraints is None:
            self._constraints = []
        self._constraints.append(value)

    def _parse_XML(self, indent, label):
        prettyprint = indent != ''

        indent = add_indent(indent)

        data = super(DataFlowDefinition, self)._to_XML(prettyprint)

        outfile = ''

        attributes = data.get('Attributes') or None

        if attributes is not None:
            outfile += f'{indent}<{label}{attributes}>'
        else:
            outfile += f'{indent}<{label}>'

        outfile += export_intern_data(data, indent)

        indent_child = add_indent(indent)
        indent_ref = add_indent(indent_child)

        if self.structure is not None:
            outfile += f'{indent_child}<{structureAbbr}:Structure>'

            if isinstance(self.structure, str):
                agencyID, id_, version = split_unique_id(self.structure)

                outfile += f'{indent_ref}<Ref id="{id_}" ' \
                           f'version="{version}" ' \
                           f'agencyID="{agencyID}" ' \
                           f'package="datastructure" class="DataStructure"/>'
            else:
                outfile += f'{indent_ref}<Ref id="{self.structure.id}" ' \
                           f'version="{self.structure.version}" ' \
                           f'agencyID="{self.structure.agencyID}" ' \
                           f'package="datastructure" class="DataStructure"/>'
            outfile += f'{indent_child}</{structureAbbr}:Structure>'

        outfile += f'{indent}</{label}>'

        return outfile



"""
    header file contains the parsers for the Header
"""
from datetime import datetime

from sdmxthon.model.base import InternationalString
from sdmxthon.model.utils import generic_setter, string_setter, bool_setter
from sdmxthon.utils.handlers import add_indent
from sdmxthon.utils.mappings import commonAbbr, structureAbbr


def _int_str_getter(attr):
    if attr is None:
        return None
    if isinstance(attr, InternationalString):
        if len(attr.items) == 0:
            return None
        if len(attr.items) == 1:
            values_view = attr.items.values()
            value_iterator = iter(values_view)
            first_value = next(value_iterator)
            return first_value['content']

        return attr.items

    return attr


def _writer_int_str(attr, word, abbr, indent, prettyprint):
    outfile = ""
    if attr is not None:
        data = attr._to_XML(f'{abbr}:{word}', prettyprint)
        for e in data:
            outfile += f'{indent}{e}'
    return outfile


def _writer_attr(attr, word, indent):
    outfile = ""
    if attr is not None:
        for e in attr:
            outfile += f'{indent}<{structureAbbr}:{word}>{e}' \
                       f'</{structureAbbr}:{word}>'

    return outfile


class Contact(object):
    """Contact provides defines the contact information about a party."""

    def __init__(self, name=None, department=None, role=None, telephone=None,
                 fax=None, X400=None, uri=None, email=None):
        self.name = name
        self.department = department
        self.role = role
        self.telephone = telephone
        self.fax = fax
        self.X400 = X400
        self.uri = uri
        self.email = email

    def __eq__(self, other):
        if isinstance(other, Contact):
            return (self.name == other.name and
                    self.uri == other.uri and
                    self.email == other.email and
                    self.fax == other.fax and
                    self.role == other.role and
                    self.telephone == other.telephone and
                    self.department == other.department)

    @property
    def name(self):
        """Name of the Contact"""
        return _int_str_getter(self._name)

    @name.setter
    def name(self, value):
        self._name = generic_setter(value, InternationalString)

    @property
    def department(self):
        """Department of the Contact"""
        return _int_str_getter(self._department)

    @department.setter
    def department(self, value):
        self._department = generic_setter(value, InternationalString)

    @property
    def role(self):
        """Role of the Contact"""
        return _int_str_getter(self._role)

    @role.setter
    def role(self, value):
        self._role = generic_setter(value, InternationalString)

    @property
    def telephone(self):
        """Telephone of the Contact"""
        if self._telephone is None:
            return None
        if len(self._telephone) == 1:
            return self._telephone[0]
        return self._telephone

    @telephone.setter
    def telephone(self, value):
        if value is None:
            self._telephone = []
        elif isinstance(value, list):
            self._telephone = value
        else:
            raise TypeError('Telephone must be a list')

    @property
    def fax(self):
        """Fax of the Contact"""
        if self._fax is None:
            return None
        if len(self._fax) == 1:
            return self._fax[0]
        return self._fax

    @fax.setter
    def fax(self, value):
        if value is None:
            self._fax = []
        elif isinstance(value, list):
            self._fax = value
        else:
            raise TypeError('Fax must be a list')

    @property
    def X400(self):
        """IPM of the Contact"""
        if self._x400 is None:
            return None
        if len(self._x400) == 1:
            return self._x400[0]

        return self._x400

    @X400.setter
    def X400(self, value):
        if value is None:
            self._x400 = []
        elif isinstance(value, list):
            self._x400 = value
        else:
            raise TypeError('X400 must be a list')

    @property
    def uri(self):
        """URI of the Contact"""
        if self._uri is None:
            return None
        if len(self._uri) == 1:
            return self._uri[0]

        return self._uri

    @uri.setter
    def uri(self, value):
        if value is None:
            self._uri = []
        elif isinstance(value, list):
            self._uri = value
        else:
            raise TypeError('URI must be a list')

    @property
    def email(self):
        """Email of the Contact"""
        if self._email is None:
            return None
        if len(self._email) == 1:
            return self._email[0]

        return self._email

    @email.setter
    def email(self, value):
        if value is None:
            self._email = []
        elif isinstance(value, list):
            self._email = value
        else:
            raise TypeError('Email must be a list')

    def to_xml(self, indent):
        outfile = ''

        prettyprint = indent != ''

        outfile += _writer_int_str(self._name, 'Name', commonAbbr, indent,
                                   prettyprint)
        outfile += _writer_int_str(self._department, 'Department',
                                   structureAbbr, indent,
                                   prettyprint)
        outfile += _writer_int_str(self._role, 'Role', structureAbbr, indent,
                                   prettyprint)

        indent = add_indent(indent)
        outfile += _writer_attr(self._telephone, 'Telephone', indent)
        outfile += _writer_attr(self._uri, 'URI', indent)
        outfile += _writer_attr(self._x400, 'X400', indent)
        outfile += _writer_attr(self._fax, 'Fax', indent)
        outfile += _writer_attr(self._email, 'Email', indent)
        return outfile


class Party(object):
    """Party defines the information which is sent about various parties
    such as senders and receivers of messages.The id attribute holds the
    identification of the party."""

    def __init__(self, id_=None, Name=None, contact=None, extensiontype_=None):
        self.id_ = id_

        if Name is None:
            self.name = []
        else:
            self.name = Name

        if contact is None:
            self.contact = []
        else:
            self.contact = contact

        self.extensiontype_ = extensiontype_

    def __eq__(self, other):
        if isinstance(other, Party):
            return (self.name == other.name and
                    self.contact == other.contact and
                    self.id_ == other.id_ and
                    self.extensiontype == other.extensiontype)

    @property
    def name(self):
        """Name of the Party"""
        return self._Name

    @name.setter
    def name(self, value):
        if value is None:
            self._Name = []
        elif isinstance(value, list):
            self._Name = value
        else:
            raise TypeError('Name must be a list')

    @property
    def contact(self):
        """Contact of the Party"""
        return self._contact

    @contact.setter
    def contact(self, value):
        if value is None:
            self._contact = []
        elif isinstance(value, list):
            self._contact = value
        else:
            raise TypeError('Contact must be a list')

    @property
    def id_(self):
        """ID of the Sender/Receiver"""
        return self._id

    @id_.setter
    def id_(self, value):
        self._id = value

    @property
    def extensiontype(self):
        """Dim_type attribute"""
        return self._extensiontype_

    @extensiontype.setter
    def extensiontype(self, value):
        self._extensiontype_ = value


class Sender(Party):
    """Sender extends the basic party structure to add an optional time
    zone declaration."""

    def __init__(self, id_=None, name=None, contact=None, timezone=None):
        super(Sender, self).__init__(id_, name, contact)
        self._timezone = timezone

    def __eq__(self, other):
        if isinstance(other, Sender):
            return self.timezone == other.timezone

    @staticmethod
    def _factory(*args_, **kwargs_):
        """Factory Method of Sender"""
        return Sender(*args_, **kwargs_)

    @property
    def timezone(self):
        """Timezone of the Sender"""
        return self._timezone

    @timezone.setter
    def timezone(self, value):
        self._timezone = value


class Header(object):
    """Header defines the basis for all message headers."""

    def __init__(self, ID=None, Test=False, Prepared=None, sender=None,
                 Receiver=None, Name=None, Structure=None,
                 DataProvider=None, DataSetAction=None, DataSetID=None,
                 Extracted=None, ReportingBegin=None,
                 ReportingEnd=None, EmbargoDate=None, Source=None):
        self._ID = ID
        self._Test = Test
        self._Prepared = Prepared
        self._Sender = sender

        if Receiver is None:
            self._Receiver = []
        else:
            self._Receiver = Receiver

        if Name is None:
            self._Name = []
        else:
            self._Name = Name

        if Structure is None:
            self._structure = {}
        else:
            self._structure = Structure

        self._dataProvider = DataProvider
        self._DataSetAction = DataSetAction

        if DataSetID is None:
            self._DataSetID = []
        else:
            self._DataSetID = DataSetID

        if isinstance(Extracted, str):
            initvalue_ = datetime.strptime(Extracted, '%Y-%m-%dT%H:%M:%S')
        else:
            initvalue_ = Extracted

        self._Extracted = initvalue_
        self._ReportingBegin = ReportingBegin
        self._ReportingEnd = ReportingEnd

        if isinstance(EmbargoDate, str):
            initvalue_ = datetime.strptime(EmbargoDate, '%Y-%m-%dT%H:%M:%S')
        else:
            initvalue_ = EmbargoDate

        self._EmbargoDate = initvalue_

        self._Source = Source

    def __eq__(self, other):
        if isinstance(other, Header):
            return (self.name == other.name and self.id_ == other.id_ and
                    self.test == other.test and self.sender == other.sender and
                    self.receiver == other.receiver and
                    self.structure == other.structure and
                    self.data_provider == other.data_provider and
                    self.dataset_action == other.dataset_action and
                    self.datasetID == other.datasetID and
                    self.extracted == other.extracted and
                    self.reporting_begin == other.reporting_begin and
                    self.reporting_end == other.reporting_end and
                    self.embargo_date == other.embargo_date and
                    self.source == other.source)

        return False

    @property
    def id_(self):
        """Identifier of the Message"""
        return self._ID

    @id_.setter
    def id_(self, value):
        self._ID = string_setter(value)

    @property
    def test(self):
        """Sets if the Message is for testing purposes"""
        return self._Test

    @test.setter
    def test(self, value):
        self._Test = bool_setter(value)

    @property
    def prepared(self):
        """Datetime of the preparation of the Message"""
        return self._Prepared

    @prepared.setter
    def prepared(self, value):
        self._Prepared = generic_setter(value, datetime)

    @property
    def sender(self):
        """Sender of the Message"""
        return self._Sender

    @sender.setter
    def sender(self, value):
        self._Sender = generic_setter(value, Sender)

    @property
    def receiver(self):
        """Receiver of the Message"""
        return self._Receiver

    @receiver.setter
    def receiver(self, value):
        if value is None:
            self._Receiver = []
        elif isinstance(value, list):
            self._Receiver = value
        elif isinstance(value, Party):
            self._Receiver = [value]
        else:
            raise TypeError('Receiver must be a list')

    @property
    def name(self):
        """Name of the Message"""
        return self._Name

    @name.setter
    def name(self, value):
        if value is None:
            self._Name = []
        elif isinstance(value, list):
            self._Name = value
        else:
            raise TypeError('Name must be a list')

    @property
    def structure(self):
        """Structures of the data in the Message"""
        return self._structure

    @structure.setter
    def structure(self, value):
        if value is None:
            self._structure = {}
        elif isinstance(value, dict):
            self._structure = value
        else:
            raise TypeError('Structure must be a dict')

    @property
    def data_provider(self):
        """Data provider of the Message"""
        return self._dataProvider

    @data_provider.setter
    def data_provider(self, value):
        self._dataProvider = value

    @property
    def dataset_action(self):
        """Action of the Dataset"""
        return self._DataSetAction

    @dataset_action.setter
    def dataset_action(self, value):
        self._DataSetAction = string_setter(value)

    @property
    def datasetID(self):
        """ID of the Dataset"""
        return self._DataSetID

    @datasetID.setter
    def datasetID(self, value):
        if value is None:
            self._DataSetID = []
        elif isinstance(value, list):
            self._DataSetID = value
        else:
            raise TypeError('DatasetID must be a list')

    @property
    def extracted(self):
        """Datetime of the information extracted"""
        return self._Extracted

    @extracted.setter
    def extracted(self, value):
        self._Extracted = generic_setter(value, datetime)

    @property
    def reporting_begin(self):
        """Start of the reporting"""
        return self._ReportingBegin

    @reporting_begin.setter
    def reporting_begin(self, value):
        self._ReportingBegin = string_setter(value)

    @property
    def reporting_end(self):
        """End of the reporting"""
        return self._ReportingEnd

    @reporting_end.setter
    def reporting_end(self, value):
        self._ReportingEnd = string_setter(value)

    @property
    def embargo_date(self):
        """Datetime of the Embargo (only in Data Messages)"""
        return self._EmbargoDate

    @embargo_date.setter
    def embargo_date(self, value):
        self._EmbargoDate = generic_setter(value, datetime)

    @property
    def source(self):
        """Source of the Message"""
        return self._Source

    @source.setter
    def source(self, value):
        self._Source = generic_setter(value, InternationalString)
